#!/usr/bin/perl

# #######################################################################
# Network analysis tool for nanobead agglomeration simulations
# [Geyer, Born, Kraus, PRL 109 (2012) 128302]
# 
# This file is part of the brownmove tutorials
# Tihamer Geyer, 2012
# 
# This script reads the brownmove trajectory from STDIN, builds a network
# per snapshot, and prints the respective degree distribution P(k)
# to STDOUT. Required parameters are the distance threshold (typically
# 9 nm in this case) and the size of the periodic simulation box 
# (100 x 100 x 100 nm3).
# 
# For more information, see the scene defintion file "nanobeads.bsc"
# #######################################################################


# Sets up a network from a distance criterion, 
# determines degree distribution

use strict;

#  pass distance threshold and size of simulation box
my $anzArgs = @ARGV;
if ($anzArgs != 4) {
	die "Usage: $0  distance  Lx Ly Lz\n";
}


# distance cutoff
my $DMAX = shift;
my $D2MAX = $DMAX*$DMAX;

# Box dimensions
my $LX = shift;
my $LY = shift;
my $LZ = shift;


my $isTimeLine = 1;
my $zeitPunkt;
my $anzProt;
my $anzProt_saved;

# Array for the network
my @net;

# arrays for the particle coordinates
my @xpos;
my @ypos;
my @zpos;

# some helper vars
my $i1; my $i2; my $i3;
my @hilf;

# switch off buffering of stdout
select(STDOUT); $| = 1;

while(<STDIN>)
{
	# copy STDIN to STDOUT
	# print $_;
	
	@hilf = split(' ');

	if($isTimeLine == 1)
	{
		$zeitPunkt = $hilf[1];
		$anzProt = $hilf[2];
		$anzProt_saved = $anzProt;
		
		$isTimeLine = 0 if ($anzProt > 0);
	} 
	else 
	{
		$anzProt--;
		
		my $plabel = $hilf[0];
		my $num = $hilf[1];
		my $slabel = $hilf[2];
		
		$xpos[$anzProt] = $hilf[3];
		$ypos[$anzProt] = $hilf[4];
		$zpos[$anzProt] = $hilf[5];
		
		# set up network
		if ($anzProt==0)
		{
			$isTimeLine = 1 ;
			
			# clear network
			@net = ();
			my $anzLinks = 0;
			
			# check all pairwise distances
			for($i1=0; $i1 < $anzProt_saved-1; $i1++) 
			{
				for($i2=$i1+1; $i2<$anzProt_saved; $i2++) 
				{
					my $dx = $xpos[$i1] - $xpos[$i2];
					my $dy = $ypos[$i1] - $ypos[$i2];
					my $dz = $zpos[$i1] - $zpos[$i2];
					
					if($dx > 0.5*$LX)     { $dx -= $LX; }
 					elsif($dx < -0.5*$LX) { $dx += $LX; }
 					if($dy > 0.5*$LY)     { $dy -= $LY; }
 					elsif($dy < -0.5*$LY) { $dy += $LY; }
 					if($dz > 0.5*$LZ)     { $dz -= $LZ; }
 					elsif($dz < -0.5*$LZ) { $dz += $LZ; }
					
					my $dr = sqrt($dx*$dx + $dy*$dy + $dz*$dz);
					
					if($dr <= $DMAX) {
						$net[$i1][$i2] = 1;
						$net[$i2][$i1] = 1;
						$anzLinks++;
						
						# print "  found links: $i1 - $i2 with DR = $dr\n";
					}
				}
			}
			
			### Network setup done, now analyze! ###
			### Network size ###
			# print "$zeitPunkt  NL  $anzProt_saved  $anzLinks\n";
			
			### P(k) ###
			my @pk;
			my $maxK = 0;
			for($i1=0; $i1 < $anzProt_saved; $i1++) {
				my $summe = 0;
				for($i2=0; $i2 < $anzProt_saved; $i2++) {
					if(defined($net[$i1][$i2])) {$summe++; }
				}
				if($summe > 15) { $summe = 15; }
				$pk[$summe]+=1;
				if($summe > $maxK) {$maxK = $summe; }
			}
			# print "$zeitPunkt  PK  $maxK";
			for($i1=0; $i1<=15; $i1++) {
				if(defined($pk[$i1])) {
				print "  ".$pk[$i1] / $anzProt_saved;
				} else {
				print "  0";
				}
			}
			print  "\n";
			
		}
	}
}

exit;
