// This file is part of the brownmove simulation package
// 
// V1.4  Aug 2012
// 
// (C) 2003 - 2012 Tihamer Geyer
// tihamer.geyer@bioinformatik.uni-saarland.de
// 
// Check for Updates at the Project Homepage 
// http://service.bioinformatik.uni-saarland.de/brownmove



#ifndef HISHAPE_H
#define HISHAPE_H

#include <iostream>
#include <string>
#include <sstream>

#include "types.h"
#include "constants.h"
#include "vectormath.h"

#include "shape.h"

namespace brown{

	/**
	 *@author uw, tg
	 */
	
	//! This class treats the hydrodynamic interactions between the Gestalt objects.

class HiShape : public Shape {
	
	public:
	
	//! Initialize without args
	HiShape(): radius(1.0) {tensor = RotnePrager; HICutOff = 1.0e20; HIScale = 1.0;}

	//!Initialize just the Radius... 
	HiShape(double r)
	: radius(r) {tensor = RotnePrager; HICutOff = 1.0e20; HIScale = 1.0;}
	
	//! Copy constructor: Initialize with the given Interacting Body.
	HiShape(const HiShape &other);
	
	//! Destructor
	~HiShape();
	
	
	//! Does nothing at the moment, as the HiShape's position is allways centered around the Gestalt's position
	virtual void recalculate(const fullPosition &p);
	
	
	//! returns some string for testing
	virtual std::string describe();
	
	//! Copy the Gestalt's actual position and uncorrelated random and external forces to HiShape
	void initForces(const fullForce &rF, const fullForce &eF);
	
	//! Caluclates the contributions to the effective forces from two HiShapes, returns < Dij / Dii >.
	//! Periodic boundaries are faked: HI is calculated with the nearest mirror image when isPeriodic == true
	double getHiCorrections(HiShape *otherHiS, bool includeRotation, const bool isPeriodic[3], const doublevec &cellSize);
	
	//! Define the radius of the one HI sphere
	void setRadius (double d) {radius = d;}
	//! Get radius of the one HI sphere
	const double getRadius() { return(radius); }
	
	//! Define a cut-off for HI
	void setHICutOff(double c, double s) { 
		std::cerr << "HiShape::setHICutOff: setting cutoff = " << c << ",  setting scale = " << s << std::endl;
		HICutOff = c;
		HIScale = s;
	}
	
	
	//! @name Query position and forces
	//! @{
	const position &getPosition() const { return(myPosition); }
	const fullForce &getRawRandomForce() { return(rawRandomForce); }
	const fullForce &getRawExternalForce() { return(rawExternalForce); }
	//! @}
	
	//! @name Sum up HI corrections to forces
	//! @{
	void addEffExternal(const fullForce &fE) {
		doublevec_add(effExternalForce.f, fE.f);
		doublevec_add(effExternalForce.t, fE.t);
	}
	void addEffRandom(const fullForce &fE) {
		doublevec_add(effRandomForce.f, fE.f);
		doublevec_add(effRandomForce.t, fE.t);
	}
	void addCorrection(const double cs[]);
	
	//! sum up the effective forces with the correct scaling and save to totalforce
	void doEffectiveForces(fullForce &totalForce, const double &beta_prime);
	//! @}
	
	
	//! Set the tensor to use.
	// static void setTensor(void (*T) (diffusionMatrix &D, const doublevec &myPos, const double myR, const doublevec &otherPos, const double otherR, bool doesRotate)) {
	static void setTensor(void (*T) (diffusionMatrix &D, const doublevec &distV, const double myR, const double otherR, bool doesRotate, const double scaling)) {
		tensor = T;
	}
	
	//! calculate the forces and torques from the charges of this and the other HiShape
	void calculateImpact(fullForce &myForce, fullForce &otherForce, HiShape *Hi);
	
	
	//! The RotnePrager Tensor
	// static void RotnePrager(diffusionMatrix& D, const doublevec &myPos, const double myR, const doublevec &otherPos, const double otherR, bool doesRotate);
	// distV is not passed as reference, because we need a scaled copy anyway
	static void RotnePrager(diffusionMatrix& D, const doublevec &distV, const double myR, const double otherR, bool doesRotate, const double scaling);
	
	//! For testing: diagonal tensor for no HI
	// static void NoHI(diffusionMatrix& D, const doublevec &myPos, const double myR, const doublevec &otherPos, const double otherR, bool doesRotate);
	static void NoHI(diffusionMatrix& D, const doublevec &distV, const double myR, const double otherR, bool doesRotate, const double scaling);
	
	
	//! Dumps the known infos about the spheres in browndef format
	void dumpStructure();

	
	
	private:
	//! @name Parameters of the hydrodynamic sphere. This is (currently) allways centered around the Gestalt's center
	//! @{
	//
	//! actual Position
	doublevec myPosition;
	//! Radius
	double radius;
	//! Potentially a cut-off and a scaling factor
	double HICutOff, HIScale;
	//! @}
	
	//! @name Forces and factors for the HI corrections
	//! @{
	//
	//! uncorrelated random and external forces
	fullForce rawRandomForce, rawExternalForce;
	//! contributions to the effective HI-corrected random and external forces (no prefactors and no own parts)
	fullForce effRandomForce, effExternalForce;
	//! \beta_ik calculated from epsilon
	double beta_ik;
	//! temp. storage for the normalization Ci factors
	double Ci_sum[6];
	//! @}
	
	//! Pointer to the HI Tensor, default = Rotne-Prager.
	//! Note: distV is not passed as reference, because we need a scaled copy anyway.
	// static void (*tensor)(diffusionMatrix& D, const doublevec &myPos, const double myR, const doublevec &otherPos, const double otherR, bool doesRotate);
	static void (*tensor)(diffusionMatrix& D, const doublevec &distV, const double myR, const double otherR, bool doesRotate, const double scaling);
	
	//! Evaluate the dyadic product of vector dist with itself, put result into matrix dyad.
	static void setupDyadMat(const doublevec& dist, tripleMatrix& dyad );
	
};

}

#endif

