// This file is part of the brownmove simulation package
// 
// V1.4  Aug 2012
// 
// (C) 2003 - 2012 Tihamer Geyer
// tihamer.geyer@bioinformatik.uni-saarland.de
// 
// Check for Updates at the Project Homepage 
// http://service.bioinformatik.uni-saarland.de/brownmove



#ifndef WALL_H
#define WALL_H

#include "interactingBody.h"
#include "cup.h"
#include "moleculeacceptor.h"
#include "moleculeinjector.h"

#include "types.h"
#include "vectormath.h"


namespace brown{
	
	class Molecule;
	
	/** The Wall: Teachers ...!
	    Contains the Gestalt and can not move.
	 *  @author jni, tg
	 */
	class Wall :public InteractingBody{
		public: 
		enum WALLTYPES {REFLECTIVE_WALL, RESERVOIR_WALL, GRID_WALL, REAL_WALL, PERIODIC_WALL};
		
		//! The simple constructor of a wall for building templates, etc. 
		//! A new wall is built with its normal in x-direction, i.e., in the y-z-plane
		Wall(string name, long int num);
		//! Copy constructor
		Wall(const Wall &templateWall);
		//! Destructor
		~Wall();
		
		//! Place wall at given full position, extends teleport from interacting body
		void teleport(const fullPosition &p);
		
		//! "We are exactly ... where ...?"
		void setMyCup(Cup *theC) {
			std::cerr << "Wall::setMyCup says Hello!" << std::endl;
			
			myCup = theC;
			
			if(myAcceptor) { myAcceptor->setCup(myCup); }
			if(myInjector) { myInjector->setCup(myCup); }
		}

		
		//! @name Injectors and Acceptors
		//! @{
		//
		//! Set and initialize an acceptor
		void setAcceptor(MoleculeAcceptor *ma){
		    myAcceptor = ma;
		    myAcceptor->setCoordSystem(&myPosition.pos, &myNormal, &myYAxis, &myZAxis);
		    myAcceptor->setCup(myCup);
		}
		
		//! Set and initialize an injector
		void setInjector(MoleculeInjector *mi){
		    myInjector = mi;
		    myInjector->setCoordSystem(&(myPosition.pos), &myNormal, &myYAxis, &myZAxis);
		    myInjector->setCup(myCup);
		}
		
		//! Inquiries
		MoleculeInjector *getInjector() { return(myInjector); }
		
		//! More diagnostics
		MoleculeAcceptor *getAcceptor() { return(myAcceptor); }
		//! @}
		
		//! Let the wall know its identity (so that it knows which complexes to have)
		void setWallType(WALLTYPES wt){ myWallType = wt; }
		//! I am a ... wall!
		WALLTYPES getWallType(void){ return(myWallType); }
		
		const doublevec &getNormal() const { return myNormal; }
		const doublevec &getYAxis() const { return myYAxis; }
		const doublevec &getZAxis() const { return myZAxis; }
		
		void setPeriodicShift(const double &dx, const double &dy, const double &dz) {
			doublevec_set(periodicDX, dx, dy, dz);
		}
		
		//! Initiate the interaction between the wall and a molecule.
		//! Handles all interactions and insertion/deletion.
		//! Molecule has to be deleted, if return value = true
		virtual bool talkToMolecule(Molecule *m, bool isPeriodic[3], const doublevec &cellSize);
		
		//! Trigger the injector to insert new molecules into the cup
		void insertNewMolecules();
		
		//! Structure of vector of saved positions plus double for
		//! reservoir
		//! @todo Do we still need MoleculeShelf in the wall?
		//	struct moleculeShelf {
		//	    vector<fullPosition> position;
		//	    double concentration;
		//	};
		//! get size of savePos-vector
		// 	int getSavePosSize() { return (savePos.size()); }
		//! add a new vector of save positions, return index of latest entry
		// 	int addNewSavePos() {
		// 	    int gross = savePos.size();
		// 	    moleculeShelf newPos;
		// 	    savePos.push_back(newPos);
		//	    return(gross);
		//	}
		
		
		protected:  // as a heritage to obstacle
		
		//! walls have to know their cup to take out and put in Molecules
		Cup *myCup;
		//! where to put the molecules (an acceptor can handle multiple particle types simultaneously)
		MoleculeAcceptor* myAcceptor;
		//! Entry for new molucules (e.g. from a reservoir or a density grid, multiple types simultaneously)
		MoleculeInjector *myInjector;
		//! Type of the Wall: REFLECTIVE_WALL, PERIODIC_WALL, RESERVOIR_WALL, GRID_WALL, REAL_WALL, PERIODIC_WALL
		WALLTYPES myWallType;
		// 	//! space to save positions of molecules of the acceptors/injectors
		// 	vector< moleculeShelf > savePos;
		
		//! Shift for periodic walls: particle is displaced by this vector when bumping into the wall
		position periodicDX;
		
		//! coordinate system of the wall (normal vector = X-Axis), used instead of rotation
		//! @todo move normal (and wall fixed coordiante system) down into the shapes and update them as the EstatShape's chargePos
		direction myNormal;
		direction myYAxis;
		direction myZAxis;
	};
}
#endif
