// This file is part of the brownmove simulation package
// 
// V1.4  Aug 2012
// 
// (C) 2003 - 2012 Tihamer Geyer
// tihamer.geyer@bioinformatik.uni-saarland.de
// 
// Check for Updates at the Project Homepage 
// http://service.bioinformatik.uni-saarland.de/brownmove



#include "cup.h"

#include "wall.h"
#include "molecule.h"


using namespace std;

namespace brown{
	
	Cup::Cup(){
		Current_Time = 0;
	}
	
	// Destructor
	Cup::~Cup(){
		// delete active molecules
		for(list<Molecule*>::iterator mlc = Active_Molecules.begin(); mlc != Active_Molecules.end(); ++mlc) {
			delete (*mlc);
		}
		Active_Molecules.clear();
		Doomed_Molecules.clear();
		resMap.clear();
		
		// delete walls
		clearWalls();
	}
	
	void Cup::initialize(){
		// revert to non-periodic setup
		doublevec_set(unitCellSize, -1.0);
		isPeriodic[0] = isPeriodic[1] = isPeriodic[2] = false;
		
		// delete active molecules
		for(int molCount = Active_Molecules.size() ; molCount > 0; molCount--)
		removeMolecule(0);
		
		Next_Molecule_Id = 1;
		Active_Molecules.clear();
		Doomed_Molecules.clear();
		resMap.clear();
		
		// reset max numbers of vdw and sticky parameters
		// Vdw_Array_Max_Index = 0L;
		stickyArray_Max_Index = 0L;
	}
	
	
	string Cup::describe() const{
		return "A Cup";
	}
	
	
	void Cup::dumpMolecules() {
		list<Molecule*>::iterator iMol;
		
		std::cout << "# Dumping molecules\n#" << std::endl;

		for(iMol = Active_Molecules.begin(); iMol != Active_Molecules.end(); iMol++) {
			(*iMol)->dumpStructure();
		}
		
		std::cout << "\n# Dump finished" << std::endl;
	}
	
	
	
	void Cup::addMolecule(const Molecule &oldMolecule) {
		// std::cerr << "Cup::addMolecule: before:  Next_Molecule_Id = " << Next_Molecule_Id << std::endl;
		Molecule *newMolecule = new Molecule(oldMolecule, 0L);
		Next_Molecule_Id += newMolecule->setIDNumber(Next_Molecule_Id);
		newMolecule->setInsertionTime(Current_Time);
		newMolecule->setvdWParameterArray(Vdw_Array, stickyArray);
		Active_Molecules.push_back(newMolecule);
		// std::cerr << "Cup::addMolecule: after:   Next_Molecule_Id = " << Next_Molecule_Id << std::endl;
	}
	
	long int Cup::getNextMoleculeID() {
		return(Next_Molecule_Id++);
	}
	
	void Cup::removeMolecule(Molecule* aMolecule){
		delete aMolecule;
		Active_Molecules.remove(aMolecule);
	}
	
	void Cup::setTimestep(double a) {
		timestep = a;
	}
	

	void Cup::removeMolecule(int pos){
		list<Molecule*>::iterator from, to;
		
		if((pos < Active_Molecules.size()) && (pos >= 0))
		{
			from = Active_Molecules.begin();
			for( ;pos>0; pos--) {
				++from;
			}
			to = from;
			++to;
			delete (*from);
			Active_Molecules.erase(from, to);
		}
		else
			cerr << "*** Cup::removeMolecule: trying to remove molecule beyond list bounds ***" << endl;
	}
	
	void Cup::doomMolecule(Molecule *m) {
		m->setTakeoutTime(Current_Time);
		Doomed_Molecules.push_back(m);
	}
	
	
	
	void Cup::removeDoomedMols() {
		int n;
		moleculeID mid;
		whereWhen inStamp, outStamp;
		
		if((n = Doomed_Molecules.size()) > 0)
		{
			for(list<Molecule*>::iterator mlc = Doomed_Molecules.begin(); mlc != Doomed_Molecules.end(); ++mlc) {
				mid = (*mlc)->getid();
				inStamp = (*mlc)->getInsertionTime();
				outStamp = (*mlc)->getTakeoutTime();
				
				if((inStamp.where != "") && (outStamp.where != "")) {
					cout << "--- Removing: "<< mid.mclass << " id " <<mid.mnumber<<"  in: "<< inStamp.when << " at " << inStamp.where << "  out: "<<outStamp.when << " at " << outStamp.where << endl;
				}
				delete (*mlc);
				Active_Molecules.remove(*mlc);
			}
		}
		
		Doomed_Molecules.clear();
	}
	
	
	
	// wall handling methods
	void Cup::addWall(const Wall &oldWall){
		Wall *newWall;
		
		std::cerr << "Cup::addWall started" << std::endl;
		
		newWall = new Wall(oldWall);
		
		std::cerr << "Cup::addWall made new wall" << std::endl;
		
		newWall->setMyCup(this);
		
		position pp;
		doublevec_set(pp, newWall->getPosition());
		cout << "*** Cup::addWall: wall pos = "<<pp[0]<<" "<<pp[1]<<" "<<pp[2]<<endl;
		
		if(newWall->getInjector()) {
			newWall->getInjector()->resolveResNames(&resMap);
		}
		if(newWall->getAcceptor()) {
			newWall->getAcceptor()->resolveResNames(&resMap);
		}
		
		My_Walls.push_back(newWall);
	}
	
	
	
	
	void Cup::clearWalls() {
		for(vector<Wall*>::iterator wl = My_Walls.begin(); wl != My_Walls.end(); ++wl)
		delete (*wl);
		
		My_Walls.clear();
	}

		
	
	void Cup::doOneStep(const timestamp &it)
	{
		// tell molecules to initialize random forces
		randomForcesOnMolecules();
		
		// std::cerr << "Cup::doOneStep: randomForces done" << std::endl;
		
		// calculate forces mol<->mol and mol<->wall
		interactMoleculeswithMolecules();
		
		// std::cerr << "Cup::doOneStep: interactMoleculeswithMolecules done" << std::endl;
		
		// check for collisions with walls
		interactMoleculeswithWalls();
		
		// std::cerr << "Cup::doOneStep: interactMoleculeswithWalls done" << std::endl;
		
		// calculate and apply Langevin corrections to the forces
		correctDamping();
		
		// std::cerr << "Cup::doOneStep: correctDamping done" << std::endl;
		
		// calculate and apply hydrodynamic corrections
		interactHi();
		
		// std::cerr << "Cup::doOneStep: interactHi done" << std::endl;
		
		// get displacements and move the molecules
		tellMoleculesToMove();
		
		// std::cerr << "Cup::doOneStep: tellMoleculesToMove done" << std::endl;
		
		// remove molecules due to leaving the sim volume or due to reactions
		removeDoomedMols();
		
		// std::cerr << "Cup::doOneStep: removeDoomedMols done" << std::endl;
		
		// insert new molecules from the outside or to replace them in reactions
		insertNewMolecules(it);
		
		// std::cerr << "Cup::doOneStep: insertNewMolecules done" << std::endl;
	}
	
	
	void Cup::randomForcesOnMolecules()
	{
		for(list<Molecule*>::iterator mlc = Active_Molecules.begin(); mlc != Active_Molecules.end(); ++mlc) 
		{
			std::vector<Gestalt>::iterator it;
			std::vector<Gestalt> *gstV = (*mlc)->getGestalten();
			
			for(it = gstV->begin(); it != gstV->end(); it++) {
				(it)->initRandomForces();
			}
		}
	}
	 

	void Cup::interactMoleculeswithWalls()
	{
		// first loop over all molecules
		for(list<Molecule*>::iterator mlc = Active_Molecules.begin(); mlc != Active_Molecules.end(); ++mlc) 
		{
			// no interaction with fixed walls, if molecule immobilized
			// -> no forces needed, also no reaction will take place
			// Check for counter == 1 to open a window for delayed reactions
			// with bound (immobilized) molecules 
			if((*mlc)->getImmobilized() <= 1)
			{
				// then loop over all walls
				for(vector<Wall*>::iterator wl = My_Walls.begin(); wl != My_Walls.end(); ++wl) {
					(*wl)->talkToMolecule(*mlc, isPeriodic, unitCellSize);
				}
				
				// now look for any external forces
				(*mlc)->whatIsPushing();
			}
		}
	}
	
	
	void Cup::interactMoleculeswithMolecules(){
		moleculeID id1, id2;
		std::vector<Gestalt>::iterator it1, it2;
		std::vector<Gestalt> *gstV, *gstV2;
		list<Molecule*>::iterator mlc;
		
		// first loop: inside the same molecule
		for(mlc = Active_Molecules.begin(); mlc != Active_Molecules.end(); ++mlc) 
		{
			gstV = (*mlc)->getGestalten();
			
			if(gstV->size() >= 2) {
				for(it1 = gstV->begin(); it1 != (--(gstV->end())); it1++) {
					for(it2 = it1; it2 != (--(gstV->end()));) {
						it2++;
						it1->talkToGestalt(*it2, true, isPeriodic, unitCellSize);		// include check for bonds (2nd arg)
					}
				}
			}
		}
		
		// second loop: gestalten from different molecules
		if(Active_Molecules.size() >= 2)
		{
			for(list<Molecule*>::iterator mlc1 = Active_Molecules.begin(); mlc1 != (--(Active_Molecules.end())); ++mlc1) 
			{
				for(list<Molecule*>::iterator mlc2 = mlc1; mlc2 != (--(Active_Molecules.end())); ) 
				{
					++mlc2;
					
					gstV = (*mlc1)->getGestalten();
					gstV2 = (*mlc2)->getGestalten();
					
					for(it1 = gstV->begin(); it1 != gstV->end(); it1++) {
						for(it2 = gstV2->begin(); it2 != gstV2->end(); it2++) {
							it1->talkToGestalt(*it2, false, isPeriodic, unitCellSize);	// no bonds between gestalten of different molecules (2nd arg)
						}
					}
				}
			}
		}
	}
	
	
	
	void Cup::correctDamping()
	{
		list<Molecule*>::iterator mlc;
		std::vector<Gestalt> *gstV;
		std::vector<Gestalt>::iterator it1;

		for(mlc = Active_Molecules.begin(); mlc != Active_Molecules.end(); ++mlc) 
		{
			gstV = (*mlc)->getGestalten();
			
			for(it1 = gstV->begin(); it1 != gstV->end(); it1++) {
				it1->correctDamping();
			}
		}
	}
	
	
	void Cup::interactHi() 
	{
		list<Molecule*>::iterator mlc;
		std::vector<Gestalt>::iterator it1, it2;
		std::vector<Gestalt> *gstV, *gstV2;
		long int numOfHiShapes = 0;
		
		// init forces in all available hishapes, count number of hishapes
		for(mlc = Active_Molecules.begin(); mlc != Active_Molecules.end(); ++mlc) 
		{
			gstV = (*mlc)->getGestalten();
			
			// size of gestalten vector does not work: not all of them need to have an hishape
			for(it1 = gstV->begin(); it1 != gstV->end(); it1++) {
				numOfHiShapes += (it1->initHiForces());
			}
		}
		// std::cerr << "Cup::interactHi: numOfHiShapes = " << numOfHiShapes << std::endl;
		
		
		// get HI corrections, sum up epsilon = < Dij/Dii >
		double epsilon_sum = 0.0, epsilon_anz = 0.0;	// use double for counting to save the type casting later on
		
		// first loop: inside the same molecule
		for(mlc = Active_Molecules.begin(); mlc != Active_Molecules.end(); ++mlc) 
		{
			gstV = (*mlc)->getGestalten();
			
			if(gstV->size() >= 2) {
				for(it1 = gstV->begin(); it1 != (--(gstV->end())); it1++) {
					for(it2 = (it1+1); it2 != gstV->end(); it2++) {
						double help = (it1->talkHi(*it2, isPeriodic, unitCellSize));		// returns -1 when no HiShape available
						if(help >= 0.0) {
							epsilon_sum += help;
							epsilon_anz += 1.0;
						}
					}
				}
			}
		}
		
		
		// second loop: gestalten from different molecules
		if(Active_Molecules.size() >= 2)
		{
			for(list<Molecule*>::iterator mlc1 = Active_Molecules.begin(); mlc1 != (--(Active_Molecules.end())); ++mlc1) 
			{
				for(list<Molecule*>::iterator mlc2 = mlc1; mlc2 != (--(Active_Molecules.end())); ) 
				{
					mlc2++;
					
					gstV = (*mlc1)->getGestalten();
					gstV2 = (*mlc2)->getGestalten();
					
					for(it1 = gstV->begin(); it1 != gstV->end(); it1++) {
						for(it2 = gstV2->begin(); it2 != gstV2->end(); it2++) {
							double help = (it1->talkHi(*it2, isPeriodic, unitCellSize));		// returns -1 when no HiShapes available
							if(help >= 0.0) {
								epsilon_sum += help;
								epsilon_anz += 1.0;
							}
						}
					}
				}
			}
		}
		
		// total number of off-diagonal entries in upper off-diagonal part of the matrix: 
		// Trans + Rot:  36 per pair of HiShapes + 15 for each HiShape
		// Trans alone:  9 per pair of HiShapes + 3 for each HiShape
		// => take twice this number for complete matrix minus diagonal
		// num of HiShapes
		// double numOfHiShapes_calc = (1.0 + sqrt(8.0 * epsilon_anz + 1)) / 2.0;		// for comparison at debugging
		// std::cerr << "Cup::interactHi:  numOfHiShapes = " << numOfHiShapes << " -- " << numOfHiShapes_calc << std::endl;

		#ifdef HI_WITH_ROTATION
		double numOfEntries = 2.0*(numOfHiShapes * 15.0 + epsilon_anz * 36.0);
		#else
		double numOfEntries = 2.0*(numOfHiShapes * 3.0 + epsilon_anz * 9.0);		// +++ HI-Change +++
		#endif
		
		// std::cerr << "Cup::interactHi:  numOfEntries = " << numOfEntries << std::endl;
		
		// calculate beta' 
		double avg_eps = epsilon_sum / numOfEntries;
		
		// std::cerr << "Cup::interactHi: epsilon = " << avg_eps << std::endl;

		// calculate beta'
		double beta_prime = -1.0;
		if(numOfEntries > 0.0) {
			// check for eps > 0
			if(avg_eps > 0.0) {
				#ifdef HI_WITH_ROTATION
				double help = (6.0*numOfHiShapes - 1.0) * avg_eps*avg_eps - (6.0*numOfHiShapes - 2.0) * avg_eps;
				#else
				double help = (3.0*numOfHiShapes - 1.0) * avg_eps*avg_eps - (3.0*numOfHiShapes - 2.0) * avg_eps;		// +++ HI-Change +++
				#endif
				beta_prime = (1.0 - sqrt(1.0 - help)) / help;
			}
			else {
				beta_prime = 0.5;
			}
		}
		
		// std::cerr << "Cup::interactHi:  beta_prime = " << beta_prime << std::endl;
		
		// distribute beta' and sum up force contributions into gestalt's totalForce
		// in Gestalt: if no HiShape => just add random- and external forces
		for(mlc = Active_Molecules.begin(); mlc != Active_Molecules.end(); ++mlc) 
		{
			gstV = (*mlc)->getGestalten();
			
			for(it1 = gstV->begin(); it1 != gstV->end(); it1++) {
				it1->doEffectiveForces(beta_prime);
			}
		}
	}
	

	
	
	void Cup::tellMoleculesToMove()
	{
		for(list<Molecule*>::iterator mlc = Active_Molecules.begin(); mlc != Active_Molecules.end(); ++mlc) 
		{
			std::vector<Gestalt>::iterator it;
			std::vector<Gestalt> *gstV = (*mlc)->getGestalten();
			
			for(it = gstV->begin(); it != gstV->end(); it++) {
				// calculate displacement from all the force contributions and perform step
				(it)->makeBrownianStep();
			}
		}
	}

	
	void Cup::immobilizeMolecules(long int delay)
	{
		for(list<Molecule*>::iterator mlc = Active_Molecules.begin(); mlc != Active_Molecules.end(); ++mlc) 
		{
			(*mlc)->immobilize(delay);
		}
	}
	
	
	void Cup::insertNewMolecules(const timestamp &it)
	{
		// set the internal clock -- neccessary that this can be triggered externally
		Current_Time = it;

		// insert molecules from the walls
		int i = 0;
		for(vector<Wall*>::iterator wl = My_Walls.begin(); wl != My_Walls.end(); ++wl, ++i) {
			(*wl)->insertNewMolecules();
		}
	}
	
	
	void Cup::listMolecules(){
		moleculeID myID;
		int i=1;
		
		for(list<Molecule*>::iterator mlc = Active_Molecules.begin(); mlc != Active_Molecules.end(); ++mlc, ++i)
		{
			myID = (*mlc)->getid();
			cerr << "*** Cup::listMolecules: " << i << " moleculeID = " << myID.mnumber << endl;
		}
	}
	
	
	long int Cup::countMolecules(string mclass) {
		moleculeID mID;
		long int anzahl = 0;
		
		for(list<Molecule*>::iterator mlc = Active_Molecules.begin(); mlc != Active_Molecules.end(); ++mlc)
		{
			mID = (*mlc)->getid();
			if(mID.mclass == mclass) {
				++anzahl;
			}
		}
		
		return(anzahl);
	}
	
	
	
	void Cup::takeMolSnapshot(FILE *fp, double absTime)
	{
		long int counter = 0;

		// count Gestalt objects (remember: position of each Gestalt is printed)
		for(list<Molecule*>::iterator mlc = Active_Molecules.begin(); mlc != Active_Molecules.end(); ++mlc) {
			counter += (*mlc)->size();
		}

		// timeline and the tell every molecule to print its Gestalts' positions
		fprintf(fp, "timeline\t%lf\t%ld\n", absTime, counter);
		for(list<Molecule*>::iterator mlc = Active_Molecules.begin(); mlc != Active_Molecules.end(); ++mlc) {
			(*mlc)->positionToSnapshot(fp);
		}
		fflush(fp);
	}
	
	
	void Cup::takeMolSnapshot_filtered(FILE *fp, double absTime, list<string> auswahl)
	{
		moleculeID mID;
		
		fprintf(fp, "timeline\t%lf\t%ld\n", absTime, countMolecules());
		
		for(list<Molecule*>::iterator mlc = Active_Molecules.begin(); mlc != Active_Molecules.end(); ++mlc) {
			mID = (*mlc)->getid();
			
			list<string>::iterator it;
			for(it=auswahl.begin(); it!=auswahl.end(); ++it) {
				if ((*it) == mID.mclass) {
					(*mlc)->positionToSnapshot(fp);
				}
			}
		}
		fflush(fp);
	}
		
	

	void Cup::listWalls()
	{
		position p;
		direction n;
		int i=1;
		
		for(vector<Wall*>::iterator wl = My_Walls.begin(); wl != My_Walls.end(); ++wl, ++i)
		{
			doublevec_set(p, (*wl)->getPosition());
			doublevec_set(n, (*wl)->getNormal());
			cerr << "*** Cup::listWalls: "<<(*wl)->getid().mclass<<"\t"<<p[0]<<" "<<p[1]<<" "<<p[2]<<"\t  "<<n[0]<<" "<<n[1]<<" "<<n[2]<<"   "<<(*wl)->getWallType()<<endl;
			
			if((*wl)->getInjector()) {
				cerr << "     Injector:  " << ((*wl)->getInjector()->describe()) << std::endl;
			}
			if((*wl)->getAcceptor()) {
				cerr << "     Acceptor:  " << ((*wl)->getAcceptor()->describe()) << std::endl;
			}
		}
	}
	
	
	fullForce Cup::getProbeForce(string probeLabel)
	{
		//! @todo Re-enable getProbeForce (in Gestalt), filter by Protein+Gestalt labels

// 		moleculeID mID;
// 		
// 		for(list<Molecule*>::iterator mlc = Active_Molecules.begin(); mlc != Active_Molecules.end(); ++mlc) {
// 			mID = (*mlc)->getid();
// 			
// 			if (mID.mclass == probeLabel) {
// 				return((*mlc)->getTotalForce());
// 			}
// 		}
	}
	
	
	void Cup::initializeStickyArray(long int dim)
	{
		if(!(stickyArray = (stickyParameter*)(new stickyParameter[dim]))) {
			std::cerr << "Cup::initializeStickyArray: Error allocating StickyArray, exiting.." << std::endl;
			exit(-1);
		}
		
		stickyArray_Max_Index = dim;
		
		// non-initialize the elements (mark as non-initialized)
		for(long int i=0; i<stickyArray_Max_Index; i++) {
			stickyArray[i].inited = false;
		}
	}
	
	
	void Cup::setStickyParam(const long int i, const double C_radial, const double DR_radial, const double C_angular, const double DR_angular, const double viscosity)
	{
		if((i>=0) && (i<stickyArray_Max_Index)) {
			stickyArray[i].inited     = true;
			stickyArray[i].C_radial   = C_radial;
			stickyArray[i].DR_radial  = DR_radial;
			stickyArray[i].C_angular  = C_angular;
			stickyArray[i].DR_angular = DR_angular;
			stickyArray[i].viscosity  = viscosity;
		}
		else {
			std::cerr << "setStickyParam: index too large" << std::endl;
			exit(-1);
		}
	}
	

	
	void Cup::initializeVDWArray(long int dim)
	{
		// build up array structure
		if(!(Vdw_Array = (vdWParameter**)(new vdWParameter*[dim]))) {
			cerr << "Cup::initializeVDWArray: Error allocating Vdw_Array, exiting.." << endl;
			exit(-1);
		}
		for(long int i=0; i<dim; i++) {
			if(!(Vdw_Array[i] = (vdWParameter*)(new vdWParameter[dim]))) {
				cerr << "Cup::initializeVDWArray: Error allocating Vdw_Array["<<i<<"], exiting.." << endl;
				exit(-1);
			}
		}
		
		Vdw_Array_Max_Index = dim;
		
		// initialize each of the elements
		for(long int i=0; i<dim; i++) {
			for(long int k=0; k<dim; k++) {
				Vdw_Array[i][k].inited = false;
				Vdw_Array[i][k].C_12 = 0.0;
				Vdw_Array[i][k].C_6  = 0.0;
				Vdw_Array[i][k].C_3  = 0.0;
				Vdw_Array[i][k].R_0  = 1.0;		// do not initialize to zero (div. by R_0!)
				Vdw_Array[i][k].DR 	 = 0.0;
				Vdw_Array[i][k].Rmin = 1.0;
				Vdw_Array[i][k].Rmax = -1e10;	// should be large and negative so that calculation will be skipped (any distance would be larger than this init values)
			}
		}
	}
	
	
	// vdW parameters are symmetric in the index
	void Cup::setVDWParam(const long int i, const long int k, enum VDW_INTERACTION itype, const double c_12, const double c_6, const double c_3, const double r_0, const double dr, const double Ass, const double Acc, const double Lc, const double rmin, const double rmax, const double agrid)
	{
		if((i>=0) && (k>=0) && (i<Vdw_Array_Max_Index) && (k<Vdw_Array_Max_Index))
		{
			Vdw_Array[i][k].inited  = true;
			Vdw_Array[k][i].inited  = true;
			Vdw_Array[i][k].intType = itype;
			Vdw_Array[k][i].intType = itype;
			Vdw_Array[i][k].C_12  = c_12;
			Vdw_Array[k][i].C_12  = c_12;
			Vdw_Array[i][k].C_6   = c_6;
			Vdw_Array[k][i].C_6   = c_6;
			Vdw_Array[i][k].C_3   = c_3;
			Vdw_Array[k][i].C_3   = c_3;
			Vdw_Array[i][k].R_0   = r_0;
			Vdw_Array[k][i].R_0   = r_0;
			Vdw_Array[i][k].DR	  = dr;
			Vdw_Array[k][i].DR	  = dr;
			Vdw_Array[i][k].ASS	  = Ass;
			Vdw_Array[k][i].ASS	  = Ass;
			Vdw_Array[i][k].ACC	  = Acc;
			Vdw_Array[k][i].ACC	  = Acc;
			Vdw_Array[i][k].LC	  = Lc;
			Vdw_Array[k][i].LC	  = Lc;
			Vdw_Array[i][k].Rmin  = rmin;
			Vdw_Array[k][i].Rmin  = rmin;
			Vdw_Array[i][k].Rmax  = rmax;
			Vdw_Array[k][i].Rmax  = rmax;
			Vdw_Array[i][k].Agrid = agrid;
			Vdw_Array[k][i].Agrid = agrid;
		} 
		else {
			cerr << "Cup::setVDWParam: index too large" << endl;
			exit(-1);
		}
	}
	
	
	void Cup::setUnitCellSize(const double dx, const double dy, const double dz) {
		if(dx > 0.0) {
			unitCellSize[0] = dx;
			isPeriodic[0] = true;
		}
		if(dy > 0.0) {
			unitCellSize[1] = dy;
			isPeriodic[1] = true;
		}
		if(dz > 0.0) {
			unitCellSize[2] = dz;
			isPeriodic[2] = true;
		}
	}

	
	void Cup::addReservoir(std::string name, double conc, double ivol)
	{
		struct reservoir res;
		
		res.density = conc;
		res.invVol = ivol;
		resMap[name] = res;
	}
	
	const double Cup::getDensity(string resName) {
		return(resMap[resName].density);
	}

	
// 	void Cup::printForces()
// 	{
// 		list<Molecule*>::iterator it;
// 		int i =0;
// 		
// 		std::cerr << "Printing all Forces from all Molecules :" << std::endl;
// 		
// 		for(it = Active_Molecules.begin(); it != Active_Molecules.end(); ++it,++i)
// 		{	
// 			fullForce frc = (*it)->getTotalForce();
// 			std::cerr << i << ": (" << frc.f[0] << "," << frc.f[1] << "," << frc.f[2] << ")" << std::endl;
// 		}
// 		
// 		std::cerr << "Printing all Forces ended!" << std::endl;
// 	}
	
}
