// This file is part of the brownmove simulation package
// 
// V1.4  Aug 2012
// 
// (C) 2003 - 2012 Tihamer Geyer
// tihamer.geyer@bioinformatik.uni-saarland.de
// 
// Check for Updates at the Project Homepage 
// http://service.bioinformatik.uni-saarland.de/brownmove



#include "hbondshape.h"

namespace brown{
	// access to global variables
	extern double HBOND_DEPTH;
	extern double HBOND_WIDTH;
	extern double HBOND_CUTOFF;
	extern double HBOND_SPRING;

	//! (Empty) Constructor
	HBondShape::HBondShape(){
	}
	
	//! Copy constructor
	HBondShape::HBondShape(const HBondShape &other){
		hooks = other.hooks;
		actual_hooks = other.actual_hooks;
	}
	
	//! Destructor
	HBondShape::~HBondShape(){
		hooks.clear();
		actual_hooks.clear();
	}
	
	
	std::string HBondShape::describe(){
		ostringstream ausgabe("", ios::out);
		int i=0;
		
		ausgabe << "actual hbond positions";
		vector<hbondPos>::iterator mlc;
		for(mlc = actual_hooks.begin(); mlc != actual_hooks.end(); ++mlc, i++){
			ausgabe << " : "<<(*mlc).pos[0]<<" "<<(*mlc).pos[1]<<" "<<(*mlc).pos[2]<<" ("<<(*mlc).bond[0]<<" "<<(*mlc).bond[1]<<" "<<(*mlc).bond[2]<<")" ; 
		}
		ausgabe <<" :";
		for(mlc = hooks.begin(); mlc != hooks.end(); ++mlc, i++){
			ausgabe << " : "<<(*mlc).pos[0]<<" "<<(*mlc).pos[1]<<" "<<(*mlc).pos[2]<<" ("<<(*mlc).bond[0]<<" "<<(*mlc).bond[1]<<" "<<(*mlc).bond[2]<<")" ; 
		}
		return(ausgabe.str());
	}
	
	
	void HBondShape::recalculate(const fullPosition &p) 
	{
		vector<hbondPos>::iterator act = actual_hooks.begin();
		for(vector<hbondPos>::iterator mlc = hooks.begin(); mlc != hooks.end(); ++mlc, ++act)
		{
			doublevec_set((*act).pos, (*mlc).pos);
			doublevec_rotate((*act).pos, p.rot);
			positionMove((*act).pos, p.pos);
			doublevec_set((*act).dr, (*act).pos);
			doublevec_sub((*act).dr, p.pos);
			
			doublevec_set((*act).bond, (*mlc).bond);
			doublevec_rotate((*act).bond, p.rot);
			positionMove((*act).bond, p.pos);
			doublevec_set((*act).drb, (*act).bond);
			doublevec_sub((*act).drb, p.pos);
		}
	}
	
	
	void HBondShape::addHook(hbondPos hp, const position &moleculeCenter){
		direction diff;
		doublevec_set(hp.dr, hp.pos);			// position of donor/acceptor rel. to mol. center
		doublevec_sub(hp.dr, moleculeCenter);
		doublevec_set(hp.drb, hp.bond);			// position of hydrogen rel. to mol. center
		doublevec_sub(hp.drb, moleculeCenter);
		doublevec_set(diff, hp.bond);			// length of atom-hydrogen vector
		doublevec_sub(diff, hp.pos);
		hp.length = doublevec_length(diff);
		hooks.push_back(hp);					// hook it up
		actual_hooks.push_back(hp);
	}
	
	
	void HBondShape::calculateImpact(fullForce &myForce, fullForce &otherForce, HBondShape *hs, const bool isPeriodic[3], const doublevec &cellSize)
	{
		position dr, ri, dr_temp, dr_help, dr_hh, dr_hh_temp;
		double r12, rhh, prefactor;
		double myScale, otherScale;
		force df;
		torque dt;
		
		// interact all hooks with each other -- nothing happens, if there are none
		// Begin with the other's hooks -- we have at least one or we should not be here
		for(vector<hbondPos>::iterator mlc2 = hs->actual_hooks.begin(); mlc2 != hs->actual_hooks.end(); ++mlc2)
		{
			otherScale = (*mlc2).giveOrTake;	// donor has scale > 0.0
			
			// cerr <<"\n*** HBondShape::calculateImpact  mlc2:   scale = " << otherScale << "  pos = " <<(*mlc2).pos[0]<<" "<<(*mlc2).pos[1]<<" "<<(*mlc2).pos[2] << "  bond = "<<(*mlc2).bond[0]<<" "<<(*mlc2).bond[1]<<" "<<(*mlc2).bond[2] << endl;

			// Now my own hooks
			for(vector<hbondPos>::iterator mlc1 = actual_hooks.begin(); mlc1 != actual_hooks.end(); ++mlc1)
			{
				// check for donor-acceptor pair
				myScale = (*mlc1).giveOrTake;
				// if(myScale * otherScale < 0.0)
				{
					// cerr <<"*** HBondShape::calculateImpact  mlc1:   scale = " << myScale << "  pos = " <<(*mlc1).pos[0]<<" "<<(*mlc1).pos[1]<<" "<<(*mlc1).pos[2] << "  bond = "<<(*mlc1).bond[0]<<" "<<(*mlc1).bond[1]<<" "<<(*mlc1).bond[2] << endl;
					
					// distance donor-acceptor
					doublevec_set(dr, (*mlc2).pos);
					doublevec_sub(dr, (*mlc1).pos);
					
					// vector between H positions
					doublevec_set(dr_hh, (*mlc2).bond);
					doublevec_sub(dr_hh, (*mlc1).bond);
					
					double dx = 0.0;
					double dy = 0.0;
					double dz = 0.0;
					int ix = 2, iy = 2, iz = 2;
					
					double likeRepulsion = 1.0;
					if(myScale * otherScale < 0.0) {
						likeRepulsion = -1.0;
					}

					// if periodic, start with displaced hooks
					if(isPeriodic[0]) {
						dx = -cellSize[0];
						ix = 0;
					}
					if(isPeriodic[1]) {
						dy = -cellSize[1];
						iy = 0;
					}
					if(isPeriodic[2]) {
						dz = -cellSize[2];
						iz = 0;
					}
					
					while(true)
					{
						// this is for the (quasi) periodic setup: distance between the hydrogen positions
						dr_hh_temp[0] = dr_hh[0] + dx;
						dr_hh_temp[1] = dr_hh[1] + dx;
						dr_hh_temp[2] = dr_hh[2] + dx;
						
						rhh = doublevec_length(dr_hh_temp);
						
						// cerr <<"*** HBondShape::calculateImpact  dr_temp:  " << dr_temp[0] << "  " << dr_temp[1] << "  " << dr_temp[2] << "  r12 = " << r12 << endl;
						
						// check if distance between hooks is shorter than cut-off
						if(rhh <= HBOND_CUTOFF) 
						{
							// attraction between the hydrogens 
							// Note: myScale * otherScale < 0  => already contains a minus sign -- only for Donor-Acceptor-pair!!!
							prefactor = 2.0 * HBOND_DEPTH * myScale * otherScale * exp(-rhh*rhh/(HBOND_WIDTH*HBOND_WIDTH)); // includes normalization of the H-H-distance vector
							doublevec_set(df, dr_hh_temp);
							doublevec_mult(df, prefactor);
										  
							doublevec_sub(myForce.f, df);
							doublevec_add(otherForce.f, df);
							
							// cerr << "   otherForce.f = " << otherForce.f[0] << "  " << otherForce.f[1] << "  " << otherForce.f[2] << endl;
							// cerr << "   otherForce.t = " << otherForce.t[0] << "  " << otherForce.t[1] << "  " << otherForce.t[2] << endl;
							// cerr << "   myForce.f = " << myForce.f[0] << "  " << myForce.f[1] << "  " << myForce.f[2] << endl;
							// cerr << "   myForce.t = " << myForce.t[0] << "  " << myForce.t[1] << "  " << myForce.t[2] << endl;

							// calculate and add torques
							// add correction for me
							doublevec_kreuzprod(dt, (*mlc1).drb, df);
							doublevec_sub(myForce.t, dt);
							doublevec_kreuzprod(dt, (*mlc2).drb, df);
							doublevec_add(otherForce.t, dt);

							
							// repulsion between donor and acceptor
							// Note: myScale * otherScale < 0  => already contains a minus sign -- only for Donor-Acceptor-pair!!!
							dr_temp[0] = dr[0] + dx;
							dr_temp[1] = dr[1] + dy;
							dr_temp[2] = dr[2] + dz;
							
							r12 = doublevec_length(dr_temp);
							
							if(r12 <= ((*mlc1).length + (*mlc2).length)) 
							{
								prefactor = -HBOND_SPRING * myScale * otherScale * likeRepulsion;	// make it repulsive for likes
								doublevec_set(df, dr_temp);
								doublevec_mult(df, prefactor);
								
								doublevec_sub(myForce.f, df);
								doublevec_add(otherForce.f, df);
								
								doublevec_kreuzprod(dt, (*mlc1).dr, df);
								doublevec_sub(myForce.t, dt);
								doublevec_kreuzprod(dt, (*mlc2).dr, df);
								doublevec_add(otherForce.t, dt);
							}
						}
						
						
						// for the quasi periodic boundaries: displace relative displacement and redo above evaluation
						if(ix < 2) {
							ix++;
							dx += cellSize[0];
						}
						else
						{
							if(isPeriodic[0]) {
								ix = 0;
								dx = -cellSize[0];
							}
							
							if(iy < 2) {
								iy++;
								dy += cellSize[1];
							}
							else
							{
								if(isPeriodic[1]) {
									iy = 0;
									dy = -cellSize[1];
								}
								
								if(iz < 2) {
									iz++;
									dz += cellSize[2];
								}
								else
								{
									break;
								}
							}
						}
					}
				}
			}
		}
	}

	
	
	void HBondShape::dumpStructure() {
		std::cout << "\n\t\tHBondShape myHHooks" << std::endl;
		
		for(std::vector<hbondPos>::iterator it = hooks.begin(); it != hooks.end(); it++) {
			if((*it).giveOrTake > 0.0) {
				std::cout << "\t\t\tdonor   "<< (*it).pos[0]<<" "<<(*it).pos[1]<<" "<<(*it).pos[2]<<"   "<< (*it).bond[0]<<" "<<(*it).bond[1]<<" "<<(*it).bond[2]<<"   "<<(*it).giveOrTake<< std::endl;
			}
			else {
				std::cout << "\t\t\tacceptor   "<< (*it).pos[0]<<" "<<(*it).pos[1]<<" "<<(*it).pos[2]<<"   "<< (*it).bond[0]<<" "<<(*it).bond[1]<<" "<<(*it).bond[2]<<"   "<<(-(*it).giveOrTake)<< std::endl;
			}
		}
		
		std::cout << "\t\tEnd" << std::endl;
	}

	
}
