// This file is part of the brownmove simulation package
// 
// V1.4  Aug 2012
// 
// (C) 2003 - 2012 Tihamer Geyer
// tihamer.geyer@bioinformatik.uni-saarland.de
// 
// Check for Updates at the Project Homepage 
// http://service.bioinformatik.uni-saarland.de/brownmove



#include "stickyshape.h"

namespace brown{

	StickyShape::StickyShape(){
		doublevec_set(ex, 1.0, 0.0, 0.0);
		doublevec_set(ey, 0.0, 1.0, 0.0);
		doublevec_set(ez, 0.0, 0.0, 1.0);
	}
	
	StickyShape::StickyShape(const StickyShape &other){
		doublevec_set(ex, other.ex);
		doublevec_set(ey, other.ey);
		doublevec_set(ez, other.ez);

		stickies = other.stickies;
		actual_stickies = other.actual_stickies;
	}
	
	StickyShape::~StickyShape(){
		stickies.clear();
		actual_stickies.clear();
	}
	
	
	std::string StickyShape::describe(){
		ostringstream ausgabe("", ios::out);
		int i=0;
		
		ausgabe << "actual patch positions";
		vector<stickyPos>::iterator mlc;
		for(mlc = actual_stickies.begin(); mlc != actual_stickies.end(); ++mlc, i++){
			ausgabe << " : "<<(*mlc).pos[0]<<" "<<(*mlc).pos[1]<<" "<<(*mlc).pos[2] ; 
		}
		ausgabe <<" :";
		for(mlc = stickies.begin(); mlc != stickies.end(); ++mlc, i++){
			ausgabe << " : "<<(*mlc).pos[0]<<" "<<(*mlc).pos[1]<<" "<<(*mlc).pos[2] ; 
		}
		return(ausgabe.str());
	}
	
	
	void StickyShape::recalculate(const fullPosition &p) 
	{
		// rotate internal coordinate system: reset vectors and rotate
		doublevec_set(ex, 1.0, 0.0, 0.0);
		doublevec_rotate(ex, p.rot);
		doublevec_set(ey, 0.0, 1.0, 0.0);
		doublevec_rotate(ey, p.rot);
		doublevec_set(ez, 0.0, 0.0, 1.0);
		doublevec_rotate(ez, p.rot);
		
		// move around the spheres
		vector<stickyPos>::iterator act = actual_stickies.begin();
		for(vector<stickyPos>::iterator mlc = stickies.begin(); mlc != stickies.end(); ++mlc, ++act)
		{
			doublevec_set((*act).pos, (*mlc).pos);
			doublevec_rotate((*act).pos, p.rot);
			positionMove((*act).pos, p.pos);
			doublevec_set((*act).dr, (*act).pos);
			doublevec_sub((*act).dr, p.pos);
		}
	}
	
	
	void StickyShape::addPatch(stickyPos sp, const position &moleculeCenter){
		doublevec_set(sp.dr, sp.pos);
		doublevec_sub(sp.dr, moleculeCenter);
		stickies.push_back(sp);
		actual_stickies.push_back(sp);
	}
	
	
	long int StickyShape::setID(const long int currNr) {
		long int incr = 0;
		for(vector<stickyPos>::iterator mlc = stickies.begin(); mlc != stickies.end(); ++mlc) {
			(*mlc).myID = currNr + incr;
			incr++;
		}
		incr = 0;
		for(vector<stickyPos>::iterator mlc = actual_stickies.begin(); mlc != actual_stickies.end(); ++mlc) {
			(*mlc).myID = currNr + incr;
			incr++;
		}

		// std::cerr << "StickyShape::setID: initializing " << incr << " IDs starting from " << currNr << std::endl;
		return(incr);
	}
	
	
	// test for stiilAlive==false
	bool resetTester(const contactSite &cs) {
		return(!(cs.stillAlive));
	}
	
	
	void StickyShape::resetFlags() {
		for(vector<stickyPos>::iterator mlc1 = actual_stickies.begin(); mlc1 != actual_stickies.end(); ++mlc1) {
			// std::cout << "StickyShape::resetFlags: sphere " << (*mlc1).myID << " -- resetting list" << std::endl;
			(*mlc1).hits.remove_if(resetTester);
			
			// std::cout << "   new list = ";
			for(std::list<contactSite>::iterator it = (*mlc1).hits.begin(); it != (*mlc1).hits.end(); it++) {
				(*it).stillAlive = false;
				// std::cout << " " << (*it).otherID; 
			}
			// std::cout << endl;
		}
	}
	



	
	void StickyShape::calculateImpact(fullForce &myForce, fullForce &otherForce, StickyShape *ss, const bool isPeriodic[3], const doublevec &cellSize)
	{
		position dr, ri, dr_temp, touch, oldTouch1, oldTouch2, difference, helper, e12;
		double r12, R_in, R_out;
		// 		double r12, prefactor, B12, help, Rmin;
		force df;
		torque dt;
		
		// interact all sticky spheres with each other -- nothing happens, if there are no spheres
		// Begin with the other's spheres -- it may have none
		for(vector<stickyPos>::iterator mlc2 = ss->actual_stickies.begin(); mlc2 != ss->actual_stickies.end(); ++mlc2)
		{
			// cout <<"*** StickyShape::calculateImpact  mlc2 ID = "<< (*mlc2).myID <<  ":  " << (*mlc2).pos[0] << "  " << (*mlc2).pos[1] << "  " << (*mlc2).pos[2] << endl;
			
			// Now my own stickies (we wouldn't be here if there were none)
			for(vector<stickyPos>::iterator mlc1 = actual_stickies.begin(); mlc1 != actual_stickies.end(); ++mlc1)
			{
				// cout <<"  StickyShape::calculateImpact  mlc1 ID = " << (*mlc1).myID << ": " << (*mlc1).pos[0] << "  " << (*mlc1).pos[1] << "  " << (*mlc1).pos[2] << endl;
				// cout << "StickyShape:calculateImpact:  ex = " << ex[0]<<", "<<ex[1]<<", "<<ex[2] << std::endl;
				
				// calculate relative distances etc
				// for periodic images: simply imply/require that boxsize is larger than two particles
				// => the same pair will not interact more than once!!!!
				doublevec_set(dr, (*mlc2).pos);
				doublevec_sub(dr, (*mlc1).pos);
				R_in  = (*mlc1).r_in  + (*mlc2).r_in;
				R_out = (*mlc1).r_out + (*mlc2).r_out;
				
				double dx = 0.0;
				double dy = 0.0;
				double dz = 0.0;
				int ix = 2, iy = 2, iz = 2;
				
				if(isPeriodic[0]) {
					dx = -cellSize[0];
					ix = 0;
				}
				if(isPeriodic[1]) {
					dy = -cellSize[1];
					iy = 0;
				}
				if(isPeriodic[2]) {
					dz = -cellSize[2];
					iz = 0;
				}
				
				while(true)
				{
					dr_temp[0] = dr[0] + dx;
					dr_temp[1] = dr[1] + dy;
					dr_temp[2] = dr[2] + dz;
					
					
					r12 = doublevec_length(dr_temp);
					
					// cout << "StickyShape::calculateImpact: r12 = " << r12 << "  R_in = " << R_in << "  R_out = " << R_out << std::endl;
					
					// already known?
					bool itsAHit = false;
					for(std::list<contactSite>::iterator it = (*mlc1).hits.begin(); it != (*mlc1).hits.end(); it++) 
					{
						if((*it).otherID == (*mlc2).myID)
						{
							// cout << "    StickyShape::calculateImpact: sphere " << (*mlc2).myID << " already known" << std::endl;
							
							// have to identify the other's hitSite
							// cout << "    StickyShape::calculateImpact: len(list) = " << (*mlc2).hits.size() << std::endl;
							std::list<contactSite>::iterator it2;
							for(it2 = (*mlc2).hits.begin(); it2 != (*mlc2).hits.end(); it2++) {
								// cout << "      StickyShape::calculateImpact: compare otherID " << (*it2).otherID << " to myID " << (*mlc1).myID << std::endl;
								if((*it2).otherID == (*mlc1).myID) {
									break;  // value (target) of it2 will be used further down
								}
							}
							// cout << "    StickyShape::calculateImpact: other contact site's otherID = " << (*it2).otherID << std::endl;

							itsAHit = true;
							
							// known: calculate interaction when close enough
							if(r12 <= R_out) 
							{
								// cout << "      StickyShape::calculateImpact: "<<(*mlc1).myID<<" and "<<(*mlc2).myID<< " close enough => calculating interaction" << std::endl;
								
								// first myself
								doublevec_set(oldTouch1,  ex);				// reconstruct old touch position w/ effect of last step
								doublevec_mult(oldTouch1, (*it).lastPos[0]);
								doublevec_set(helper,     ey);
								doublevec_mult(helper,    (*it).lastPos[1]);
								doublevec_add(oldTouch1,  helper);
								doublevec_set(helper,     ez);
								doublevec_mult(helper,    (*it).lastPos[2]);
								doublevec_add(oldTouch1,  helper);
								
								// the other one
								doublevec_set(oldTouch2,  ss->ex);
								doublevec_mult(oldTouch2, (*it2).lastPos[0]);
								doublevec_set(helper,     ss->ey);
								doublevec_mult(helper,    (*it2).lastPos[1]);
								doublevec_add(oldTouch2,  helper);
								doublevec_set(helper,     ss->ez);
								doublevec_mult(helper,    (*it2).lastPos[2]);
								doublevec_add(oldTouch2,  helper);

								// calculate relative difference
								doublevec_set(difference, dr_temp);
								doublevec_add(difference, oldTouch2);
								doublevec_sub(difference, oldTouch1);
								
								// cout << "StickyShape::calculateImpact: difference = "<<difference[0]<<" "<<difference[1]<<" "<<difference[2]<<std::endl;
								
								// unit vector along ctr-ctr-line
								doublevec_set(e12, dr_temp);
								doublevec_mult(e12, 1.0/r12);
								
								// radial contribution
								double radialPart = doublevec_scalprod(e12, difference);
								doublevec_set(helper, e12);
								doublevec_mult(helper, radialPart);		// part of the difference vector along dr_temp
								doublevec_sub(difference, helper);		// difference is now the remaining angular part
								
								double newFact = paramsArray[(*mlc1).color].viscosity;
								// cout << "StickyShape::calculateImpact: viscosity originally = " << newFact << std::endl;

								// add up the forces -- if displacement small enough
								if(doublevec_length(helper) < paramsArray[(*mlc1).color].DR_radial) 
								{
									doublevec_mult(helper, paramsArray[(*mlc1).color].C_radial);
									doublevec_add(myForce.f, helper);
									doublevec_sub(otherForce.f, helper);
								} else {
									// std::cerr << "StickyShape::calculateImpact: radial displacement " << doublevec_length(helper) << " > DR_radial = " << paramsArray[(*mlc1).color].DR_radial << std::endl;
									newFact = 1.0;		// no restoring force, switch to new position
								}

								// handle torques
								if(doublevec_length(difference) < paramsArray[(*mlc1).color].DR_angular) 
								{
									doublevec_mult(difference, paramsArray[(*mlc1).color].C_angular);
									doublevec_kreuzprod(dt, oldTouch1, difference);
									doublevec_add(myForce.t, dt);
									doublevec_kreuzprod(dt, oldTouch2, difference);
									doublevec_sub(otherForce.t, dt);
								} else {
									// std::cerr << "StickyShape::calculateImpact: angular displacement " << doublevec_length(difference) <<" > DR_angular = " << paramsArray[(*mlc1).color].DR_angular << std::endl;
									newFact = 1.0;		// no restoring torque, switch to new position
								}
								
								// set stillAlive and update touch position when still close enough
								(*it).stillAlive = true;
								std::list<contactSite>::iterator it2;
								for(it2 = (*mlc2).hits.begin(); it2 != (*mlc2).hits.end(); it2++) {
									if((*it2).otherID == (*mlc1).myID) {
										(*it2).stillAlive = true;
										break;
									}
								}
								
								// reset touch position
								doublevec_set(touch, dr_temp); 	 			// dr_temp points from mlc1 to mlc2
								doublevec_mult(touch, (*mlc1).r_in/R_in);		// rescale by ratio of radii
								// cout << "StickyShape::calculateImpact: resetting touch1 = "<<touch[0]<<" "<<touch[1]<<" "<<touch[2]<<endl;
								// cout << "StickyShape::calculateImpact: viscosity = " << newFact << std::endl;
								// cout << "StickyShape::calculateImpact: lastPos1 = "<<(*it).lastPos[0]<<"  "<<(*it).lastPos[1]<<"  "<<(*it).lastPos[2]<<std::endl;
								double x_part = newFact * doublevec_scalprod(touch, ex);
								double y_part = newFact * doublevec_scalprod(touch, ey);
								double z_part = newFact * doublevec_scalprod(touch, ez);
								// cout << "StickyShape::calculateImpact: parts1 = "<<x_part<<"  "<<y_part<<"  "<<z_part<<std::endl;
								doublevec_mult((*it).lastPos, 1.0-newFact);					// viscosity = 0 => keep old position
								// cout << "StickyShape::calculateImpact: lastPos1 scaled = "<<(*it).lastPos[0]<<"  "<<(*it).lastPos[1]<<"  "<<(*it).lastPos[2]<<std::endl;
								doublevec_add((*it).lastPos, x_part, y_part, z_part);
								// cout << "StickyShape::calculateImpact: lastPos1 + parts1 = "<<(*it).lastPos[0]<<"  "<<(*it).lastPos[1]<<"  "<<(*it).lastPos[2]<<std::endl;

								
								// reset touch position
								doublevec_set(touch, dr_temp); 	 			// dr_temp points from mlc1 to mlc2
								doublevec_mult(touch, -(*mlc2).r_in/R_in);		// rescale by ratio of radii
								// cout << "StickyShape::calculateImpact: resetting touch2 = "<<touch[0]<<" "<<touch[1]<<" "<<touch[2]<<endl;
								// cout << "StickyShape::calculateImpact: lastPos2 = "<<(*it2).lastPos[0]<<"  "<<(*it2).lastPos[1]<<"  "<<(*it2).lastPos[2]<<std::endl;
								x_part = newFact * doublevec_scalprod(touch, ss->ex);
								y_part = newFact * doublevec_scalprod(touch, ss->ey);
								z_part = newFact * doublevec_scalprod(touch, ss->ez);
								// cout << "StickyShape::calculateImpact: parts2 = "<<x_part<<"  "<<y_part<<"  "<<z_part<<std::endl;
								doublevec_mult((*it2).lastPos, 1.0-newFact);					// viscosity = 0 => keep old position
								// cout << "StickyShape::calculateImpact: lastPos2 scaled = "<<(*it2).lastPos[0]<<"  "<<(*it2).lastPos[1]<<"  "<<(*it2).lastPos[2]<<std::endl;
								doublevec_add((*it2).lastPos, x_part, y_part, z_part);
								// cout << "StickyShape::calculateImpact: lastPos2 + parts2 = "<<(*it2).lastPos[0]<<"  "<<(*it2).lastPos[1]<<"  "<<(*it2).lastPos[2]<<std::endl;

							}  // if(r12 <= R_out)
						}
					}
					
					// not known yet: register when close enough
					if(!itsAHit) 
					{
						// cout << "    StickyShape::calculateImpact: myID = " << (*mlc1).myID <<":  sphere " << (*mlc2).myID << " not registered yet" << std::endl;
						
						// check whether the spheres are close enough
						if(r12 <= R_in) 
						{
							// cout << "StickyShape::calculateImpact: close enough => registering " << (*mlc1).myID <<" vs " << (*mlc2).myID << std::endl;
							
							struct contactSite cs;
							cs.otherID = (*mlc2).myID;
							cs.stillAlive = true;
							doublevec_set(touch, dr_temp); 	 			// dr_temp points from mlc1 to mlc2
							doublevec_mult(touch, (*mlc1).r_in/R_in);		// rescale by ratio of radii
							// cout << "StickyShape::calculateImpact: touch1 = "<<touch[0]<<" "<<touch[1]<<" "<<touch[2]<<endl;
							double x_part = doublevec_scalprod(touch, ex);
							double y_part = doublevec_scalprod(touch, ey);
							double z_part = doublevec_scalprod(touch, ez);
							doublevec_set(cs.lastPos, x_part, y_part, z_part);
							// cout << "StickyShape::calculateImpact: parts_1 = "<<cs.lastPos[0]<<" "<<cs.lastPos[1]<<" "<<cs.lastPos[2]<<endl;
							(*mlc1).hits.push_back(cs);
							
							cs.otherID = (*mlc1).myID;
							cs.stillAlive = true;
							doublevec_set(touch, dr_temp); 	 			// dr_temp points from mlc1 to mlc2
							doublevec_mult(touch, -(*mlc2).r_in/R_in);		// rescale by ratio of radii
							// cout << "StickyShape::calculateImpact: touch2 = "<<touch[0]<<" "<<touch[1]<<" "<<touch[2]<<endl;
							x_part = doublevec_scalprod(touch, ss->ex);
							y_part = doublevec_scalprod(touch, ss->ey);
							z_part = doublevec_scalprod(touch, ss->ez);
							doublevec_set(cs.lastPos, x_part, y_part, z_part);
							// cout << "StickyShape::calculateImpact: parts_2 = "<<cs.lastPos[0]<<" "<<cs.lastPos[1]<<" "<<cs.lastPos[2]<<endl;
							(*mlc2).hits.push_back(cs);
						}
					}
					
					// this part does the periodic boundary conditions
					if(ix < 2) {
						ix++;
						dx += cellSize[0];
					}
					else
					{
						if(isPeriodic[0]) {
							ix = 0;
							dx = -cellSize[0];
						}
						
						if(iy < 2) {
							iy++;
							dy += cellSize[1];
						}
						else
						{
							if(isPeriodic[1]) {
								iy = 0;
								dy = -cellSize[1];
							}
							
							if(iz < 2) {
								iz++;
								dz += cellSize[2];
							}
							else
							{
								break;
							}
						}
					}
				}
			}
		}
	}
		
		
	
	void StickyShape::dumpStructure() {
		std::cout << "\n\t\tStickyShape myPatches" << std::endl;
		
		for(std::vector<stickyPos>::iterator it = stickies.begin(); it != stickies.end(); it++) {
			std::cout << "\t\t\t"<< (*it).pos[0]<<" "<<(*it).pos[1]<<" "<<(*it).pos[2]<<"   "<<(*it).r_in<<"   " << (*it).r_out<<"   " << (*it).color<< std::endl;
		}
		
		std::cout << "\t\tEnd" << std::endl;
	}

	
}
